from pystorz.store import options
import logging

from datetime import datetime
from pystorz.internal import constants
from pystorz.store import store

log = logging.getLogger(__name__)


class MetaStore(store.Store):
    def __init__(self, data: store.Store):
        self._Store = data

    def Create(self, obj: store.Object, *opt: options.CreateOption) -> store.Object:
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        log.info("create {} {}".format(
            obj.Metadata().Kind(), obj.PrimaryKey()))

        meta = obj.Metadata()
        if isinstance(meta, store.MetaSetter):
            meta.SetIdentity(store.ObjectIdentityFactory())
            meta.SetCreated(datetime.now())
            meta.SetUpdated(obj.Metadata().Created())
            meta.SetRevision(1)

        return self._Store.Create(obj, *opt)

    def Update(
        self,
        identity: store.ObjectIdentity,
        obj: store.Object,
        *opt: options.UpdateOption,
    ) -> store.Object:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        log.info("update {}".format(identity.Path()))

        original = self._Store.Get(identity)
        # if original.Metadata().Kind() != obj.Metadata().Kind():
        #     return constants.ErrObjectIdentityMismatch

        # obj.Metadata().SetKind(original.Metadata().Kind())
        meta = obj.Metadata()
        if isinstance(meta, store.MetaSetter):
            meta.SetIdentity(original.Metadata().Identity())
            meta.SetCreated(original.Metadata().Created())
            meta.SetUpdated(datetime.now())
            meta.SetRevision(original.Metadata().Revision() + 1)

        return self._Store.Update(identity, obj, *opt)

    def Delete(self, identity: store.ObjectIdentity, *opt: options.DeleteOption):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info("delete {}".format(identity.Path()))
        return self._Store.Delete(identity, *opt)

    def Get(
        self, identity: store.ObjectIdentity, *opt: options.GetOption
    ) -> store.Object:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info("get {}".format(identity.Path()))
        return self._Store.Get(identity, *opt)

    def List(
        self, identity: store.ObjectIdentity, *opt: options.ListOption
    ) -> store.ObjectList:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info("list {}".format(identity.Path()))
        return self._Store.List(identity, *opt)



import json
import logging


log = logging.getLogger(__name__)

from typing import Callable


class CommonOptionHolder:
    def __init__(self):
        self.filter = None
        self.order_by = None
        self.order_incremental = None
        self.page_size = None
        self.page_offset = None

    def common_options(self):
        return self


class Option:
    def ApplyFunction(self) -> Callable[[CommonOptionHolder], None]:
        raise NotImplementedError
    
    def ToJson(self) -> str:
        raise NotImplementedError

    def ToDict(self) -> dict:
        raise NotImplementedError


class CreateOption(Option):
    def get_create_option(self) -> Option:
        raise NotImplementedError


class GetOption(Option):
    def get_get_option(self) -> Option:
        raise NotImplementedError


class UpdateOption(Option):
    def get_update_option(self) -> Option:
        raise NotImplementedError


class ListOption(Option):
    def get_list_option(self) -> Option:
        raise NotImplementedError


class DeleteOption(Option):
    def get_delete_option(self) -> Option:
        raise NotImplementedError


class ListDeleteOption(ListOption, DeleteOption):
    @staticmethod
    def FromJson(jsn):
        log.debug("parsing listdeleteoption: {}".format(jsn))

        data = jsn
        if isinstance(jsn, str):
            data = json.loads(jsn)

        if data["type"] == "in":
            return In(data["key"], data["values"])
        elif data["type"] == "eq":
            return Eq(data["key"], data["value"])
        elif data["type"] == "lt":
            return Lt(data["key"], data["value"])
        elif data["type"] == "lte":
            return Lte(data["key"], data["value"])
        elif data["type"] == "gt":
            return Gt(data["key"], data["value"])
        elif data["type"] == "gte":
            return Gte(data["key"], data["value"])
        elif data["type"] == "and":
            return And(*[ListDeleteOption.FromJson(f) for f in data["filters"]])
        elif data["type"] == "or":
            return Or(*[ListDeleteOption.FromJson(f) for f in data["filters"]])
        elif data["type"] == "not":
            return Not(ListDeleteOption.FromJson(data["filter"]))
        else:
            raise ValueError("Unknown type: {}".format(data["type"]))

    def ToJson(self) -> str:
        return json.dumps(self.ToDict())

    def ToDict(self) -> dict:
        copt = CommonOptionHolderFactory()
        self.ApplyFunction()(copt)
        if isinstance(copt.filter, Option):
            return copt.filter.ToDict()
        return {}

    def __str__(self) -> str:
        copt = CommonOptionHolderFactory()
        self.ApplyFunction()(copt)
        return str(copt.filter)
    

class OptionHolder:
    def common_options(self):
        raise NotImplementedError


class EqOption(ListDeleteOption):
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value
    
    def ToDict(self):
        return {"type": "eq", "key": self.key, "value": self.value}

    def __str__(self):
        return "{} = {}".format(self.key, self.value)
    

class LtOption(ListDeleteOption):
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value

    def ToDict(self):
        return {"type": "lt", "key": self.key, "value": self.value}

    def __str__(self):
        return "{} < {}".format(self.key, self.value)


class LteOption(ListDeleteOption):
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value

    def ToDict(self):
        return {"type": "lte", "key": self.key, "value": self.value}

    def __str__(self):
        return "{} <= {}".format(self.key, self.value)


class GtOption(ListDeleteOption):
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value
    
    def ToDict(self):
        return {"type": "gt", "key": self.key, "value": self.value}

    def __str__(self):
        return "{} > {}".format(self.key, self.value)


class GteOption(ListDeleteOption):
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value
        
    def ToDict(self):
        return {"type": "gte", "key": self.key, "value": self.value}

    def __str__(self):
        return "{} >= {}".format(self.key, self.value)


class AndOption(ListDeleteOption):
    def __init__(self, *filters: ListDeleteOption):
        self.filters = filters

    def ToDict(self):
        return {"type": "and", "filters": [f.ToDict() for f in self.filters]}

    def __str__(self):
        return "( {} )".format(
            " AND ".join([str(f) for f in self.filters]))


class OrOption(ListDeleteOption):
    def __init__(self, *filters: ListDeleteOption):
        self.filters = filters

    def ToDict(self):
        return {"type": "or", "filters": [f.ToDict() for f in self.filters]}

    def __str__(self):
        return "( {} )".format(
            " OR ".join([str(f) for f in self.filters]))


class NotOption(ListDeleteOption):
    def __init__(self, flt: ListDeleteOption):
        self.filter = flt

    def ToDict(self):
        return {"type": "not", "filter": self.filter.ToDict()}

    def __str__(self):
        return "( NOT {} )".format(self.filter)


class InOption(ListDeleteOption):
    def __init__(self, key: str, values: list):
        self.key = key
        self.values = values

    def ToDict(self):
        return {"type": "in", "key": self.key, "values": self.values}

    def __str__(self):
        return "({} IN ({}))".format(
            self.key, ", ".join([str(v) for v in self.values]))


def CommonOptionHolderFactory() -> CommonOptionHolder:
    return CommonOptionHolder()


class _ListDeleteOption(ListDeleteOption):
    def __init__(self, function):
        self.function = function

    def get_delete_option(self):
        return self

    def get_list_option(self):
        return self

    def ApplyFunction(self):
        return self.function


def And(*filters: ListDeleteOption) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = AndOption(*filters)

    return _ListDeleteOption(option_function)


def Or(*filters: ListDeleteOption) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = OrOption(*filters)

    return _ListDeleteOption(option_function)


def Not(filter: ListDeleteOption) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = NotOption(filter)

    return _ListDeleteOption(option_function)


def Eq(prop: str, val) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = EqOption(key=prop, value=val)

    return _ListDeleteOption(option_function)


def Lt(prop: str, val) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = LtOption(key=prop, value=val)

    return _ListDeleteOption(option_function)


def Gt(prop: str, val) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = GtOption(key=prop, value=val)

    return _ListDeleteOption(option_function)


def Lte(prop: str, val) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = LteOption(key=prop, value=val)

    return _ListDeleteOption(option_function)


def Gte(prop: str, val) -> ListDeleteOption:
    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = GteOption(key=prop, value=val)

    return _ListDeleteOption(option_function)


def In(key: str, values: list) -> ListDeleteOption:
    if not key or len(key) == 0:
        raise Exception("empty key for in filter")

    if not values or len(values) == 0:
        raise Exception("empty values for in filter")

    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.filter is not None:
            raise Exception("prop filter option already set")

        common_options.filter = InOption(key, values)

    return _ListDeleteOption(option_function)


def PageSize(ps: int) -> ListOption:
    if ps < 0:
        raise Exception("page size cannot be negative")

    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.page_size is not None:
            raise Exception("page size option has already been set")

        common_options.page_size = ps
        # logging.info(f"pagination size option {ps}")

    return _ListDeleteOption(option_function)


def PageOffset(po: int) -> ListOption:
    if po < 0:
        raise Exception("page offset cannot be negative")

    def option_function(options: OptionHolder):
        common_options = options.common_options()
        if common_options.page_offset is not None:
            raise Exception("page offset option has already been set")

        common_options.page_offset = po

    return _ListDeleteOption(option_function)


def Order(field, ascending=True):
    def f(options):
        common_options = options.common_options()
        if common_options.order_by is None:
            common_options.order_by = field
            common_options.order_incremental = ascending
        else:
            raise Exception("order by option has already been set")

    return _ListDeleteOption(f)



 

class _Schema(store.SchemaHolder):
	def __init__(self, objects):
		self.objects = objects
	
	def ObjectForKind(self, kind) -> store.Object:
		{% for r in data["resources"] %}
		
		if kind == "{{r.name}}":
			return {{r.name}}Factory()
		elif kind == "{{r.IdentityPrefix()}}":
			return {{r.name}}Factory()

		{% endfor %}
		
		raise Exception(constants.ErrNoSuchObject)
	
	def types(self):
		return self.objects

def Schema():
	objects = [
		{% for r in data["resources"] %}
		"{{r.name}}",
		{% endfor %}
	]

	return _Schema(objects)

 


import os
import shutil
import black
import logging
from io import StringIO

from jinja2 import Template

from pystorz.mgen.loader import load_model, Resource, Struct, Prop, typeDefault
from pystorz.store import utils

log = logging.getLogger(__name__)
log.debug('loading builder.py...')


def Generate(*models) -> None:
    structs = []
    resources = []

    for model in models:
        log.debug(f"Loading model {model}...")
        
        s, r = load_model(model)
        
        structs.extend(s)
        resources.extend(r)

    imports = [
        "import json",
        "from pystorz.internal import constants",
        # "from pystorz.store import utils",
        "from pystorz.store import store",
        "from datetime import datetime",
        "from typing import Type",
    ]

    b = StringIO()

    b.write(render("mgen/templates/imports.py", {"imports": imports}))

    b.write(compileStructs(structs))
    b.write(compileResources(resources))

    res = b.getvalue().replace("&#34;", "\"")

    # refactor and format python code
    res = reformat_python_code(res)

    targetDir = "generated"

    if os.path.exists(targetDir):
        # delete the directory with contents
        shutil.rmtree(targetDir)

    utils.export_file(targetDir, "model.py", res)


def compileResources(resources: list[Resource]) -> str:
    b = StringIO()

    for r in resources:
        log.debug(f"Compiling resource {r.name}...")

        props = [
            Prop(
                "Meta",
                "store.Meta",
                "metadata",
                f'store.MetaFactory("{r.name}")')
        ]

        if r.external:
            props.append(Prop(
                "External",
                r.external,
                "external"))

        if r.internal:
            props.append(Prop(
                "Internal",
                r.internal,
                "internal"))

        s = Struct(
            r.name,
            "store.Object",
            props,
        )

        b.write(compileStruct(s))
        b.write(render("mgen/templates/meta.py", r))

    b.write(render("mgen/templates/schema.py", {"resources": resources}))

    return b.getvalue()


def compileStructs(structs: list[Struct]) -> str:
    b = StringIO()

    for s in dependencyOrder(structs):
        b.write(compileStruct(s))
    
    return b.getvalue()


def dependencyOrder(structs: list[Struct]) -> list[Struct]:
    all_structs = {}
    dependencies = {}
    for s in structs:
        all_structs[s.name] = s
        dependencies[s.name] = set()
    
    known_types = ["string", "int", "float", "bool", "datetime"]

    for s in structs:
        for p in s.properties:
            if p.IsArray():
                elem_type = p.SubType()
                if elem_type not in known_types:
                    dependencies[s.name].add(elem_type)
                    
                continue

            if p.IsMap():
                # map[string]bla
                elem_type = p.SubType()
                key_type = p.type[4:].split("]")[0]
                if key_type not in known_types:
                    raise Exception(f"Invalid type {key_type} in {p.type}. Please use one of {known_types} as map key type")

                val_type = p.type[4:].split("]")[1]
                if val_type not in known_types:
                    dependencies[s.name].add(val_type)
                
                continue

            if p.type not in known_types:
                dependencies[s.name].add(p.type)
    
    ordered = []
    while len(ordered) < len(structs):
        did_stuff = False
        for s in structs:
            if s in ordered:
                continue
            
            if len(dependencies[s.name]) == 0:
                ordered.append(s)
                for k in dependencies.keys():
                    if s.name in dependencies[k]:
                        dependencies[k].remove(s.name)
                        did_stuff = True
        
        if not did_stuff and len(ordered) < len(structs):
            log.debug(f"dependencies: {dependencies}")
            log.debug(f"ordered: {[ o.name for o in ordered]}")
            log.debug(f"lengths: {len(dependencies)} {len(ordered)} {len(structs)}")

            raise Exception("Circular dependency detected")

    return ordered


def compileStruct(s: Struct) -> str:
    log.debug(f"Compiling struct {s.name}...")

    b = StringIO()
    methods = []

    s.properties = addDefaultPropValues(s.properties)

    parent = s.implements

    for p in s.properties:
        if p.name != "Meta":
            # methods.append(f"{p.name}(self) -> {p.StrippedType()}")
            methods.append((f"{p.name}(self)", p.StrippedType()))

        if p.name != "Meta" and p.name != "External" and p.name != "Internal":
            # methods.append(f"Set{p.name}(self, val: {p.StrippedType()})")
            methods.append((f"Set{p.name}(self, val: {p.StrippedType()})", ""))

        if p.name == "External":
            parent = "store.ExternalHolder"
    
    b.write(
        render(
            "mgen/templates/interface.py",
            {
                'name': s.name,
                'methods': methods,
                'implements': parent
            }))

    if parent in ["store.Object", "store.ExternalHolder"]:
        b.write(render("mgen/templates/clone.py", s))

    b.write(render("mgen/templates/structure.py", s))
    b.write(render("mgen/templates/unmarshall.py", s))

    if parent in ["store.Object", "store.ExternalHolder"]:
        b.write(render("mgen/templates/cloneimp.py", s))

    return b.getvalue()


def render(rpath: str, data) -> str:
    path = os.path.join(utils.runtime_dir(), rpath)

    with open(path, 'r') as f:
        content = f.read()

    t = Template(content)
    return t.render(data=data)


def addDefaultPropValues(props: list[Prop]) -> list[Prop]:
    res = []

    for p in props:
        if len(p.default) > 0:
            res.append(p)
            continue

        res.append(Prop(p.name, p.type, p.json, typeDefault(p.type)))

    return res


def reformat_python_code(code_str):
    try:
        # Reformat the code string using the parsed AST.
        formatted_code = black.format_str(code_str, mode=black.FileMode())

        return formatted_code

    except Exception as e:
        log.error("failed to reformat code:", e)
        return code_str



import os
import json
import time
import logging
import pathlib
import base64

from urllib.parse import quote, unquote

from jsonpath import JSONPath
from datetime import datetime

log = logging.getLogger(__name__)

# import gabs

from pystorz.internal import constants
from pystorz.store import store


# class _MetaHolder:
#     def __init__(self):
#         self.Metadata = None
#         self.External = None
#         self.Internal = None


def unmarshal_object(
    body, schema: store.SchemaHolder, kind: str
) -> store.Object:
    resource = schema.ObjectForKind(kind)
    resource.FromJson(body)
    return resource


def object_kind(data) -> str:
    if "metadata" not in data:
        return ""
    
    if "kind" not in data["metadata"]:
        return ""
    
    return data["metadata"]["kind"]


def pps(string: str) -> str:
    return pp(json.loads(string))


def pp(obj) -> str:
    return json.dumps(obj, indent=4)


def timestamp() -> str:
    return time.strftime("%Y-%m-%dT%H:%M:%S.%fZ", time.gmtime())


def serialize(mo: store.Object) -> str:
    if mo is None:
        raise Exception(constants.ErrObjectNil)
    return json.dumps(mo, default=lambda x: x.to_dict())


def object_path(obj: store.Object, path: str) -> object:
    # print("object_path: {} {}".format(obj, path))
    data = obj.ToDict()
    # print(pp(data))
    ret = JSONPath("$.{}".format(path)).parse(data)
    if not ret or len(ret) == 0:
        # print("object_path: no result for {} in {}".format(path, data))
        return None
    return ret[0]


def export_file(target_dir: str, name: str, content: str) -> None:
    pathlib.Path(target_dir).mkdir(parents=True, exist_ok=True)
    target_file = os.path.join(target_dir, name)
    with open(target_file, "w") as f:
        f.write(content)


def runtime_dir() -> str:
    return str(pathlib.Path(__file__).parent.parent.absolute())


def encode_string(string):
    return str(string).replace("'", "$%#")
    # encoded_message = base64.b64encode(string.encode('utf-8'))  # Encode the message as base64 bytes
    # return str(encoded_message, 'utf-8')  # Convert the bytes to a string


def decode_string(soup):
    return soup.replace("$%#", "'")
    # encoded_message = bytes(soup, 'utf-8')  # Convert the soup to bytes
    # return base64.b64decode(encoded_message).decode('utf-8')  # Decode the base64 bytes and convert to string


# encoding = 'ascii'
encoding = 'utf-8'
def base64_encode_string(string):
    # encoded_message = base64.b64encode(string.encode('utf-8'))  # Encode the message as base64 bytes
    # return str(encoded_message, 'utf-8')  # Convert the bytes to a string
    encoded_message = base64.b64encode(string.encode(encoding))  # Encode the message as base64 bytes
    return quote(str(encoded_message, encoding))  # Convert the bytes to a string


def base64_decode_string(soup):
    encoded_message = bytes(unquote(soup), encoding)  # Convert the soup to bytes
    return base64.b64decode(encoded_message).decode(encoding)  # Decode the base64 bytes and convert to string



import logging
import sqlite3
from pystorz.sql.store import SqlStore
from pystorz.sql.adapter import SQLAdapter

log = logging.getLogger(__name__)


def SqliteStoreFactory(schema, path):
    def connector():
        return _SqliteAdapter(path)

    return SqlStore(schema, connector)


class _SqliteAdapter(SQLAdapter):
    def __init__(self, path):
        super().__init__()
        self._path = path

    def connect(self):
        return sqlite3.connect(self._path)



from pystorz.store import options


class _HeaderOption(
    options.GetOption,
    options.CreateOption,
    options.UpdateOption,
    options.DeleteOption,
    options.ListOption,
):
    def __init__(self, function):
        self.function = function

    def ApplyFunction(self):
        return self.function

    def get_create_option(self) -> options.Option:
        return self

    def get_delete_option(self) -> options.Option:
        return self

    def get_get_option(self) -> options.Option:
        return self

    def get_update_option(self) -> options.Option:
        return self

    def get_list_option(self) -> options.Option:
        return self

    def get_header_option(self) -> options.Option:
        return self


def Header(key: str, val: str) -> _HeaderOption:
    def function(opts: options.OptionHolder):
        rest_opts = opts  # Assume that the `opts` argument is of type `RestOptions`
        if len(key.split(" ")) > 1:
            raise ValueError(f"invalid header name [{key}]")
        
        rest_opts.headers[key] = val

    return _HeaderOption(function)



import uuid
import json

from datetime import datetime
from pystorz.internal import constants
from pystorz.store import options


def datetime_parse(dtstr) -> datetime:
    return datetime.strptime(dtstr, constants.DATETIME_FORMAT)


def datetime_string(dt) -> str:
    return dt.strftime(constants.DATETIME_FORMAT)


class ObjectIdentity:
    def __init__(self, id: str):
        self.id_ = id

    def FromString(self, str):
        self.id_ = str

    def __str__(self) -> str:
        return self.id_

    def Path(self) -> str:
        if '/' in self.id_:
            tokens = self.id_.split('/')
            return f"{tokens[0].lower()}/{tokens[1]}"
        else:
            return f"id/{self}"

    def IsId(self) -> bool:
        return '/' not in self.id_

    def Type(self) -> str:
        return self.Path().split('/')[0]

    def Key(self) -> str:
        tokens = self.Path().split('/')
        if len(tokens) > 1:
            return tokens[1]
        else:
            return ""

    def __eq__(self, other):
        if isinstance(other, ObjectIdentity):
            return self.id_ == other.id_
        elif isinstance(other, str):
            return self.id_ == other

        return False

    def __hash__(self):
        return hash(self.id_)

    def __len__(self):
        return len(self.id_)


class Meta:
    def Kind(self) -> str:
        raise NotImplementedError()

    def Identity(self) -> ObjectIdentity:
        raise NotImplementedError()

    def Created(self) -> datetime:
        raise NotImplementedError()

    def Updated(self) -> datetime:
        raise NotImplementedError()

    def Revision(self) -> int:
        raise NotImplementedError()

    def ToJson(self) -> str:
        raise NotImplementedError()

    def FromDict(self, d: dict):
        raise NotImplementedError()

    def ToDict(self) -> str:
        raise NotImplementedError()


class MetaSetter:
    def SetKind(self, kind: str):
        raise NotImplementedError()

    def SetIdentity(self, identity: ObjectIdentity):
        raise NotImplementedError()

    def SetCreated(self, created: datetime):
        raise NotImplementedError()

    def SetUpdated(self, updated: datetime):
        raise NotImplementedError()

    def SetRevision(self, revision: int):
        raise NotImplementedError()


class _MetaWrapper(Meta, MetaSetter):
    def __init__(self, kind):
        self.revision_ = 0
        self.kind_ = kind
        self.identity_ = ObjectIdentityFactory()
        self.created_ = ""
        self.updated_ = ""

    def Kind(self) -> str:
        return self.kind_

    def Created(self) -> datetime:
        return datetime_parse(self.created_)

    def Updated(self) -> datetime:
        return datetime_parse(self.updated_)

    def Identity(self) -> ObjectIdentity:
        return self.identity_

    def Revision(self) -> int:
        return self.revision_

    def SetKind(self, kind: str) -> None:
        self.kind_ = kind

    def SetIdentity(self, identity: ObjectIdentity) -> None:
        self.identity_ = identity

    def SetCreated(self, created: datetime) -> None:
        self.created_ = datetime_string(created)

    def SetUpdated(self, updated: datetime) -> None:
        self.updated_ = datetime_string(updated)

    def SetRevision(self, revision: int) -> None:
        self.revision_ = revision

    def ToJson(self) -> str:
        return json.dumps(self.ToDict())

    def ToDict(self) -> dict:
        return {
            "kind": self.Kind(),
            "identity": str(self.Identity()),
            "created": self.created_,
            "updated": self.updated_,
            "revision": self.revision_,
        }

    def FromDict(self, d: dict) -> None:
        self.SetKind(d["kind"])
        self.SetIdentity(ObjectIdentityFactory())
        self.Identity().FromString(d["identity"])
        self.created_ = d["created"]
        self.updated_ = d["updated"]
        self.revision_ = d["revision"]


def MetaFactory(kind: str) -> Meta:
    return _MetaWrapper(kind)


class Object:

    def __init__(self):
        raise Exception("Object is an interface")

    def Metadata(self) -> Meta:
        raise Exception("Object is an interface")

    def Clone(self):
        raise Exception("Object is an interface")

    def ToJson(self) -> str:
        raise Exception("Object is an interface")

    def FromJson(self, jstr):
        raise Exception("Object is an interface")

    def FromDict(self, dict) -> Exception:
        raise Exception("Object is an interface")

    def ToDict(self) -> dict:
        raise Exception("Object is an interface")

    def PrimaryKey(self) -> str:
        raise Exception("Object is an interface")


class ExternalHolder(Object):
    def SetExternal(self, obj: object):
        pass

    def External(self) -> object:
        pass


class ObjectList(list[Object]):
    pass


def ObjectIdentityFactory() -> ObjectIdentity:
    id = str(uuid.uuid1())
    id = id.replace("-", "")

    return ObjectIdentity(id)


class Store:
    def Get(self, identity: ObjectIdentity, *options: options.GetOption) -> Object:
        raise Exception("Object is an interface")

    def List(self, identity: ObjectIdentity, *options: options.ListOption) -> ObjectList:
        raise Exception("Object is an interface")

    def Create(self, obj: Object, *options: options.CreateOption) -> Object:
        raise Exception("Object is an interface")

    def Update(self, identity: ObjectIdentity, obj: Object, *options: options.UpdateOption) -> Object:
        raise Exception("Object is an interface")

    def Delete(self, identity: ObjectIdentity, *options: options.DeleteOption):
        raise Exception("Object is an interface")


class SchemaHolder:
    def ObjectForKind(self, kind: str) -> Object:
        raise Exception("Object is an interface")

    def Types(self) -> list[str]:
        raise Exception("Object is an interface")


class StoreJsonDecoder(json.JSONDecoder):

    def __init__(self, schema: SchemaHolder):
        super().__init__(object_hook=self.object_hook)
        self.schema = schema

    def object_hook(self, dct):
        # find the class
        class_name = dct["meta"]["kind"]
        instance = self.schema.ObjectForKind(class_name)
        if instance is None:
            raise Exception(f"cannot find kind: {class_name}")

        return instance.FromJson(dct)




class {{ data["name"] }}({{ data["implements"] }}):

	def __init__(self):
		raise Exception("cannot initialize like this. use the factory method")
	
	def ToDict(self):
		raise Exception("not implemented")

	def FromDict(self, data):
		raise Exception("not implemented")

{% for method, ret in data["methods"] %}
{% if ret|length > 0 %}
	def {{ method }} -> {{ ret }}:
{% else %}
	def {{ method }}:
{% endif %}
		raise Exception("not implemented")

{% endfor %}



import json
import logging

# import traceback
import cherrypy

from pystorz.internal import constants
from urllib.parse import urlparse, parse_qs

from pystorz.store import store
from pystorz.store import options
from pystorz.store import utils

from pystorz.rest.internals import InternalStore

from flask import Flask, request


# from urllib.parse import parse_qs, urlparse


log = logging.getLogger(__name__)


FilterArg = "pf"
IncrementalArg = "inc"
PageSizeArg = "ps"
PageOffsetArg = "po"
OrderByArg = "ob"


ActionGet = "GET"
ActionCreate = "POST"
ActionUpdate = "PUT"
ActionDelete = "DELETE"


def _handle_exceptions(e):
    # traceback.print_stack()
    # print("exception: {}".format(e))

    error_code = 500
    msg = str(e)
    if msg == constants.ErrNoSuchObject:
        error_code = 404
    if msg == constants.ErrInvalidMethod:
        error_code = 405

    return _error_response(error_code, msg)


def _json_response(code: int, data):
    try:
        ret = json.dumps(data)
        log.debug("""json response: 
    {}""".format(utils.pp(data)))

        return ret, code, {"Content-Type": "application/json"}
    except Exception as e:
        log.debug("json error: {}".format(e))
        return _error_response(500, str(e))


def _error_response(code: int, message: str):
    return _json_response(code, {"error": message})


def _make_id_handler(stor, schema, exposed):
    def handler(id):
        iddentifier = store.ObjectIdentity(id)
        existing = stor.Get(iddentifier)

        robject = None
        if existing is not None:
            kind = existing.Metadata().Kind()
            obj_methods = exposed[kind]
            if obj_methods is None or request.method not in obj_methods:
                return _error_response(405, constants.ErrInvalidMethod)

            try:
                robject = utils.unmarshal_object(request.data, schema, kind)
            except Exception as e:
                log.debug("unmarshal error: {}".format(e))

        return _handle_path(stor, iddentifier, robject)

    return handler


def _handle_path(
    stor: store.Store, identity: store.ObjectIdentity, object):
    log.info("handle path {}".format(identity.Path()))
    log.info("method {}".format(request.method))

    ret = None
    if request.method == ActionGet:
        try:
            ret = stor.Get(identity)
        except Exception as e:
            return _error_response(404, str(e))
    elif request.method == ActionCreate:
        try:
            ret = stor.Create(object)
        except Exception as e:
            return _error_response(409, str(e))
    elif request.method == ActionUpdate:
        try:
            ret = stor.Update(identity, object)
        except Exception as e:
            return _error_response(409, str(e))
    elif request.method == ActionDelete:
        try:
            stor.Delete(identity)
        except Exception as e:
            return _error_response(404, str(e))

    if ret is not None:
        return _json_response(200, ret.ToDict())

    return _json_response(200, {})


def _make_object_handler(
    stor: store.Store, schema: store.SchemaHolder, t: str, methods: list
):
    def handler(pkey: str):
        log.info("handle object {} {} {}".format(pkey, t, request.method))

        id = store.ObjectIdentity("{}/{}".format(t.lower(), pkey))

        robject = None
        try:
            robject = utils.unmarshal_object(request.data, schema, t)
        except Exception as e:
            log.debug("unmarshal error: {}".format(e))

        if request.method not in methods:
            return _error_response(405, constants.ErrInvalidMethod)

        return _handle_path(stor, id, robject)

    return handler


def _make_type_handler(
    stor: store.Store,
    schema: store.SchemaHolder,
    t: str,
    methods: list
):
    def handler():
        log.info("handle type {} {} (allowed {})".format(t, request.method, methods))

        if request.method not in methods:
            return _error_response(405, constants.ErrInvalidMethod)

        if request.method == ActionGet or request.method == ActionDelete:
            url = urlparse(request.url)
            query_params = parse_qs(url.query)
            log.debug("query: {}".format(query_params))

            opts = []
            if request.data is not None and len(request.data) > 0:
                o = options.ListDeleteOption.FromJson(
                    request.data.decode("utf-8"))
                
                opts.append(o)
                log.debug("filter: {}".format(str(o)))

            # if FilterArg in query_params:
            #     fa = query_params[FilterArg]
            #     o = options.ListDeleteOption.FromJson(fa[0])
                
            #     opts.append(o)
            #     log.debug("filter: {}".format(str(o)))

            if PageSizeArg in query_params:
                ps = int(query_params[PageSizeArg][0])
                opts.append(options.PageSize(ps))

            if PageOffsetArg in query_params:
                ps = int(query_params[PageOffsetArg][0])
                opts.append(options.PageOffset(ps))

            if OrderByArg in query_params:
                ascending = True
                if IncrementalArg in query_params:
                    ascending = query_params[IncrementalArg][0] == "true"

                opts.append(options.Order(query_params[OrderByArg][0], ascending))

            try:
                if request.method == ActionGet:
                    ret = stor.List(store.ObjectIdentity(t.lower() + "/"), *opts)
                    return _json_response(200, [r.ToDict() for r in ret])                
                else:
                    ret = stor.Delete(store.ObjectIdentity(t.lower() + "/"), *opts)
                    return _json_response(200, {})
            except Exception as e:
                return _error_response(400, str(e))

        elif request.method == ActionCreate:
            try:
                robject = utils.unmarshal_object(request.data, schema, t)
            except Exception as e:
                log.debug("unmarshal error: {}".format(e))
                raise Exception(constants.ErrInvalidRequest)

            return _handle_path(stor, store.ObjectIdentity(t + "/"), robject)

        return _error_response(405, constants.ErrInvalidMethod)

    return handler


class Expose:
    def __init__(self, kind: str, *actions: str):
        self.Kind = kind
        self.Actions = actions


class Server:
    def __init__(self, schema, thestore, *to_expose: Expose):
        self.Schema = schema
        self.Store = InternalStore(schema, thestore)

        accepted_actions = set([ActionGet, ActionCreate, ActionUpdate, ActionDelete])

        exposed = {}
        for e in to_expose:
            for a in e.Actions:
                if a not in accepted_actions:
                    raise Exception("invalid action: {}".format(a))

            exposed[e.Kind] = e.Actions

        self.app = Flask(__name__)
        self.app.register_error_handler(Exception, _handle_exceptions)
        
        self._register_handlers(exposed)

        @self.app.before_request
        def _log_request_info():
            log.debug("Request: {} {}".format(request.method, request.url))
            log.debug("Headers: {}".format(request.headers))
            log.debug("Body: {}".format(request.data))


    def Serve(self, host: str, port: int):
        # launch a flask server to serve the html
        log.info("serving on {}:{}".format(host, port))

        # self.app.run(host, port)

        cherrypy.tree.graft(self.app, "/")
        cherrypy.config.update(
            {
                "server.socket_host": host,
                "server.socket_port": port,
            }
        )

        cherrypy.config.update(
            {
                "log.access_file": "",  # Disable CherryPy's default access log
                # 'log.screen': False,    # Disable logging to the console
            }
        )

        # Enable access logging
        cherrypy.log.access_log = log
        cherrypy.engine.start()

        return self.Stop
    
    def Stop(self):
        try:
            # cherrypy.engine.stop()
            cherrypy.engine.exit()
        except Exception as e:
            log.error("error stopping server: {}".format(e))
        
        log.info("server stopped...")

    def Join(self):
        try:
            cherrypy.engine.block()
        except KeyboardInterrupt:
            self.Stop()
        
    def _register_handlers(self, exposed):
        self._register_handler(
            "/id/<id>",
            "id_handler",
            _make_id_handler(self.Store, self.Schema, exposed),
            methods=[ActionGet, ActionCreate, ActionUpdate, ActionDelete],
        )

        for k, v in exposed.items():
            methods = list(v)

            self._register_handler(
                f"/{k.lower()}/<pkey>",
                f"{k}_handler",
                _make_object_handler(self.Store, self.Schema, k, methods),
                methods=methods,
            )

            type_handler = _make_type_handler(self.Store, self.Schema, k, methods)

            self._register_handler(
                f"/{k.lower()}", f"{k}_type_handler1", type_handler, methods=methods
            )

            self._register_handler(
                f"/{k.lower()}/", f"{k}_type_handler2", type_handler, methods=methods
            )

    def _register_handler(self, path: str, name: str, func, methods: list):
        log.info("registering {} at {} with {}".format(name, path, methods))

        self.app.add_url_rule(path, name, func, methods=methods)



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='pystorz',
    version='0.4.4',
    author='wazofski',
    description='Python package for the Storz object store framework.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/wazofski/pystorz',
    packages=[
        "pystorz",
        "pystorz/internal",
        "pystorz/meta",
        "pystorz/mgen",
        "pystorz/mgen/templates",
        "pystorz/rest",
        "pystorz/sql",
        "pystorz/store",
        "pystorz/router",
    ],
    package_data={'': ['*.py']},
    # package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        "PyYAML",
        "jinja2",
        "black",
        "jsonpath-python",
        "cherrypy",
        "requests",
        "mysql-connector-python",
        "sqlparse",
        # "pysqlite3",
    ],
    extras_require={
        'dev': [
            # 'pytest -v test_mgen.py -cov',
            # 'pytest -v -k "thestore" test_common.py -cov',
            # 'flake8',
        ]
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires=">=3.6",
)



	
	def FromJson(self, jstr):
		data = json.loads(jstr)
		return self.FromDict(data)


	def ToJson(self):
		return json.dumps(self.ToDict())


	def ToDict(self):
		data = {}
		{% for prop in data.properties %}
		{% if prop.IsArray() %}
		rawList = []
		for v in self.{{prop.name}}_:
			{% if prop.IsComplexType() %}
			rawList.append(v.ToDict())
			{% else %}
			rawList.append(v)
			{% endif %}

		data["{{prop.json}}"] = rawList
		{% elif prop.IsMap() %}
		rawSubmap = {}
		for k, v in self.{{prop.name}}_.items():
			{% if prop.IsComplexType() %}
			rawSubmap[k] = v.ToDict()
			{% else %}
			rawSubmap[k] = v
			{% endif %}

		data["{{prop.json}}"] = rawSubmap
		{% else %}
		
		{% if prop.IsComplexType() %}
		# if self.{{prop.name}}_ is not None:
		data["{{prop.json}}"] = self.{{prop.name}}_.ToDict()
		{% else %}
		data["{{prop.json}}"] = self.{{prop.name}}_
		{% endif %}
	
		{% endif %}
		{% endfor %}
		return data
	

	def FromDict(self, data):
		for key, rawValue in data.items():
			if rawValue is None:
				continue
			
			{% for prop in data.properties %}
			if key == "{{prop.json}}":
				{% if prop.IsArray() %}
				res = {{ prop.default }}

				for rw in rawValue:
					ud = {{ prop.ComplexTypeValueDefault() }}
					{% if prop.IsComplexType() %}
					ud.FromDict(rw)
					{% else %}
					ud = rw
					{% endif %}
					res.append(ud)

				self.{{prop.name}}_ = res
				{% elif prop.IsMap() %}
				res = {{ prop.default }}
				
				for rk, rw in rawValue.items():
					ud = {{prop.ComplexTypeValueDefault()}}
					{% if prop.IsComplexType() %}
					ud.FromDict(rw)
					{% else %}
					ud = rw
					{% endif %}
					res[rk] = ud

				self.{{prop.name}}_ = res
				{% else %}
				{% if prop.IsComplexType() %}
				self.{{prop.name}}_.FromDict(rawValue)
				{% else %}
				self.{{prop.name}}_ = rawValue
				{% endif %}
				
				{% endif %}
			{% endfor %}
		
	



	def Metadata(self):
		return self.Meta_

	def PrimaryKey(self):
		return str(self.{{data.primary_key}})


def {{data.name}}Identity(pkey):
	return store.ObjectIdentity("{{ data.IdentityPrefix() }}/" + pkey )

{{data.name}}KindIdentity = store.ObjectIdentity("{{ data.IdentityPrefix() }}/")

{{data.name}}Kind = "{{ data.name }}"



from pystorz.store import utils

def test_mgen():
    from generated.model import WorldFactory, WorldKind, Schema, NestedWorldFactory
    
    world = WorldFactory()

    # Test factory method
    assert world is not None
    assert world.External() is not None
    assert world.Internal() is not None

    # Test setters and getters
    world.External().SetName("abc")
    assert world.External().Name() == "abc"

    world.External().Nested().SetAlive(True)
    assert world.External().Nested().Alive() is True

    world.External().Nested().SetCounter(10)
    assert world.External().Nested().Counter() == 10

    world.Internal().SetDescription("qwe")
    assert world.Internal().Description() == "qwe"

    # Test metadata
    assert world.Metadata().Kind() == WorldKind

    # Test deserialization
    world.External().Nested().SetCounter(10)
    world.External().Nested().SetAlive(True)
    world.External().Nested().SetAnotherDescription("qwe")
    world.External().SetName("abc")
    world.Internal().SetDescription("qwe")
    world.Internal().SetList([NestedWorldFactory(), NestedWorldFactory()])

    world.Internal().SetMap({
        "a": NestedWorldFactory(),
        "b": NestedWorldFactory(),
    })

    world.Internal().Map()["a"].SetL1([False, False, True])

    data = world.ToJson()
    # print(utils.pps(data))
    
    newWorld = WorldFactory()
    newWorld.FromJson(data)
    
    # print(utils.pps(newWorld.ToJson()))

    assert newWorld.External().Nested().Alive() is True
    assert newWorld.External().Nested().Counter() == 10
    assert newWorld.External().Name() == "abc"
    assert newWorld.Internal().Description() == "qwe"
    assert len(newWorld.Internal().List()) == 2
    data2 = newWorld.ToJson()
    assert data == data2

    # Test schema
    schema = Schema()
    obj = schema.ObjectForKind(str(world.Metadata().Kind()))
    assert obj is not None
    anotherWorld = obj.Clone()
    assert anotherWorld is not None

    # Test cloning
    world.External().Nested().SetCounter(10)
    world.External().Nested().SetAlive(True)
    world.External().Nested().SetAnotherDescription("qwe")
    world.External().SetName("abc")
    world.Internal().SetDescription("qwe")
    world.Internal().SetList([NestedWorldFactory(), NestedWorldFactory()])

    world.Internal().SetMap({
        "a": NestedWorldFactory(),
        "b": NestedWorldFactory(),
    })

    world.Internal().Map()["a"].SetL1([False, False, True])

    newWorld = world.Clone()
    assert newWorld.External().Nested().Alive() is True
    assert newWorld.External().Nested().Counter() == 10
    assert newWorld.External().Nested().AnotherDescription() == "qwe"
    assert newWorld.External().Name() == "abc"
    assert newWorld.Internal().Description() == "qwe"
    assert len(newWorld.Internal().List()) == 2



import logging
import mysql.connector
from pystorz.sql.store import SqlStore
from pystorz.sql.adapter import SQLAdapter


log = logging.getLogger(__name__)


def MySqlStoreFactory(schema, host, port, user, password, database):
    def connector():
        return _MySQLAdapter(host, port, user, password, database)

    return SqlStore(schema, connector)


class _MySQLAdapter(SQLAdapter):
    def __init__(self, host, port, username, password, database):
        super().__init__()

        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._database = database

    def connect(self):
        return mysql.connector.connect(
            host=self._host,
            port=self._port,
            user=self._username,
            password=self._password,
            database=self._database
        )



import logging
log = logging.getLogger(__name__)


class SQLAdapter:
    def __init__(self):
        self._connection = None
        self._cursor = None

    def connect(self):
        raise NotImplementedError()

    def connection(self):
        if not self._connection:
            self._cursor = None
            self._connection = self.connect()

        return self._connection

    def cursor(self, retry=True):
        try:
            if not self._cursor:
                connection = self.connection()
                self._cursor = connection.cursor()

            return self._cursor
        except Exception as err:
            if not retry:
                raise err

        self._connection = None
        self._cursor = None
        return self.cursor(False)

    def close(self):
        try:
            if self._cursor:
                self._cursor.close()

            if self._connection:
                self._connection.close()
        except Exception as err:
            log.error("close", err)

        self._connection = None
        self._cursor = None

    def execute(self, query, params=None, retry=True):
        try:
            cursor = self.cursor()
            if params:
                return cursor.execute(query, params)
            else:
                return cursor.execute(query)
        except Exception as err:
            if not retry:
                raise err

        self._cursor = None
        self._connection = None
        return self.execute(query, params, False)

    def fetchall(self, retry=True):
        try:
            return self.cursor().fetchall()
        except Exception as err:
            if not retry:
                raise err

        self._cursor = None
        self._connection = None
        return self.fetchall(False)

    def fetchone(self, retry=True):
        try:
            return self.cursor().fetchone()
        except Exception as err:
            if not retry:
                raise err

        self._cursor = None
        self._connection = None
        return self.fetchone(False)

    def commit(self, retry=True):
        try:
            return self.connection().commit()
        except Exception as err:
            log.info(f"exception: {retry}")
            if not retry:
                raise err

        self._cursor = None
        self._connection = None
        return self.commit(False)

    def rollback(self, retry=True):
        try:
            return self.connection().rollback()
        except Exception as err:
            if not retry:
                raise err

        self._cursor = None
        self._connection = None
        return self.rollback(False)
